package tools.descartes.prisma.core.agent;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import org.eclipse.emf.cdo.CDOAdapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

import edu.kit.ipd.descartes.mm.resourcetype.ResourceType;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceTypeRepository;
import edu.kit.ipd.descartes.mm.resourcetype.ResourcetypePackage;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapterFactory;
import tools.descartes.prisma.core.adapter.ModelSkeletonDelta;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.core.owner.OwnershipManager;
import tools.descartes.prisma.core.scopes.AgentScope;
import tools.descartes.prisma.core.scopes.SystemGlobalScope;
import tools.descartes.prisma.model.scope.Agent;
import tools.descartes.prisma.model.scope.DelegatedObject;
import tools.descartes.prisma.model.scope.ScopeConfiguration;
import tools.descartes.prisma.model.scope.ScopePackage;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class AgentController {

	private class DelegationRepositoryAdapter implements CDOAdapter {

		private Agent target;

		@Override
		public void notifyChanged(Notification msg) {
			if (msg.getFeatureID(
					ScopeConfiguration.class) == ScopePackage.SCOPE_CONFIGURATION__DELEGATED_OBJECTS) {
				if (msg.getEventType() == Notification.ADD) {
					DelegatedObject delegatedElement = (DelegatedObject) msg.getNewValue();
					notifyAddDelegatedElement(delegatedElement.getObject(),
							delegatedElement.getAdditionalProperties().map());
				} else if (msg.getEventType() == Notification.ADD_MANY) {
					List<?> newValues = (List<?>) msg.getNewValue();
					for (Object curValue : newValues) {
						DelegatedObject delegatedElement = (DelegatedObject) curValue;
						notifyAddDelegatedElement(delegatedElement.getObject(),
								delegatedElement.getAdditionalProperties().map());
					}
				} else if (msg.getEventType() == Notification.REMOVE) {
					DelegatedObject delegatedElement = (DelegatedObject) msg.getOldValue();
					notifyRemoveDelegatedElement(delegatedElement.getObject(),
							delegatedElement.getAdditionalProperties().map());
				} else if (msg.getEventType() == Notification.REMOVE_MANY) {
					List<?> oldValues = (List<?>) msg.getOldValue();
					for (Object curValue : oldValues) {
						DelegatedObject delegatedElement = (DelegatedObject) curValue;
						notifyRemoveDelegatedElement(delegatedElement.getObject(),
								delegatedElement.getAdditionalProperties().map());
					}
				}
			}
		}

		@Override
		public Agent getTarget() {
			return target;
		}

		@Override
		public void setTarget(Notifier newTarget) {
			target = (Agent) target;
		}

		@Override
		public boolean isAdapterForType(Object type) {
			return (type instanceof Agent);
		}
	}

	private static final int MAX_RETRIES = 3;

	private Agent agent;
	private ModelRepository repository;
	private final ScheduledExecutorService workerThread = Executors.newScheduledThreadPool(1);
	private final String scopePath;
	private final String id;
	private AgentScope scope;
	private OwnershipManager ownership;
	private ModelSkeletonAdapterFactory adapterFactory;

	private DelegationProvider delegationProvider = new DelegationProvider() {
		@Override
		public boolean needsDelegation(EObject object) {
			return false;
		}

		@Override
		public Map<String, String> getAdditionalProperties(EObject object) {
			return Collections.emptyMap();
		}
	};

	private volatile boolean running = false;

	public AgentController(ModelRepository repository, String scopePath, String id) {
		this.repository = repository;
		this.scopePath = scopePath;
		this.id = id;
	}

	public Agent getAgent() {
		return agent;
	}

	public AgentScope getContainerScope() {
		return scope;
	}

	public OwnershipManager getOwnershipManager() {
		return ownership;
	}

	public void setDelegationProvider(DelegationProvider provider) {
		this.delegationProvider = provider;
	}

	public Future<?> start() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (!running) {
					startAsync();
				}
				running = true;
			}
		});
	}

	public Future<?> stop() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (running) {
					stopAsync();
				}
				running = false;
			}
		});
	}

	public void apply(ModelSkeleton skeleton) {
		int retries = 0;

		while (retries <= MAX_RETRIES) {

			try (Transaction transaction = getModelRepository().createTransaction()) {

				replacePublicResources(skeleton);

				scope.synchronizeModelSkeleton(this, transaction, skeleton);

				boolean success = true;

				try {
					transaction.commit();
				} catch (CommitException ex) {
					System.err.println("Commit conflict. Retry...");
					success = false;
				}

				List<ModelSkeletonAdapter> adapters = new LinkedList<>();
				for (EObject curObj : skeleton.eContents()) {
					ModelSkeletonAdapter curAdapter = adapt(curObj, ModelSkeletonAdapter.class);
					adapters.add(curAdapter);
				}
				// IMPORTANT: commit after all model skeleton adapters have been
				// synchronized and merged.
				for (ModelSkeletonAdapter curAdapter : adapters) {
					ModelSkeletonDelta delta = curAdapter.commitFinished(success);

					if (success) {
						// Delegate the new objects.
						for (EObject curObject : delta.getNewObjects()) {
							if (delegationProvider.needsDelegation(curObject)) {
								scope.delegate(curObject, delegationProvider.getAdditionalProperties(curObject));
							}
						}
						// Undelegate the removed objects.
						for (EObject curObject : delta.getRemovedObjects()) {
							if (delegationProvider.needsDelegation(curObject)) {
								// TODO:
							}
						}
					}
				}

				// The commit was successful -> just return to caller, otherwise
				// retry
				if (success) {
					return;
				}
			}
			retries++;
		}

		// Maximum number of retries reached. Failure
		System.err.println("Maximum number of retries reached. Commit was not successful.");
	}

	protected ScheduledFuture<?> schedule(Runnable command, long period, TimeUnit unit) {
		return workerThread.scheduleAtFixedRate(command, 0, period, unit);
	}

	protected void startAsync() {
		repository.connect();
		SystemGlobalScope systemScope = getModelRepository().getSystemScope();
		systemScope.create();
		this.scope = (AgentScope) getModelRepository().getScope(scopePath);
		this.scope.create();
		this.agent = scope.addAgent(id);
		this.adapterFactory = new ModelSkeletonAdapterFactory(repository, this);
		this.ownership = new OwnershipManager(getModelRepository(), scope.getScopeConfiguration(), this.agent);

		// Deliver the notifications for the delegated objects which were
		// registered before this agent started.
		for (DelegatedObject existingDelegation : this.scope.getScopeConfiguration().getDelegatedObjects()) {
			notifyAddDelegatedElement(existingDelegation.getObject(),
					existingDelegation.getAdditionalProperties().map());
		}

		// Listen for changes in the delegated elements list in our scope so
		// that we can react to delegation requests we are interested in.
		this.scope.getScopeConfiguration().eAdapters().add(new DelegationRepositoryAdapter());
	}

	protected void stopAsync() {
		try {
			scope.removeAgent(agent);
		} finally {
			repository.disconnect();
		}
	}

	public ModelRepository getModelRepository() {
		return repository;
	}

	protected void notifyAddDelegatedElement(EObject newElement, Map<String, String> additionalProperties) {
	}

	protected void notifyRemoveDelegatedElement(EObject oldElement, Map<String, String> additionalProperties) {
	}

	protected void checkRunning() {
		if (!running) {
			throw new IllegalStateException("The agent is not running.");
		}
	}

	@SuppressWarnings("unchecked")
	public <T> T adapt(EObject obj, Class<T> adapterClass) {
		return (T) adapterFactory.adapt(obj, adapterClass);
	}

	private void replacePublicResources(ModelSkeleton skeleton) {
		// Resource types (e.g., CPU, HDD, DELAY) which are defined globally are
		// not included in the model skeleton.
		// We need to find these references in the model skeleton and connect
		// them with the instances in the repository.
		TreeIterator<EObject> iterator = skeleton.eAllContents();
		ResourceTypeRepository publicResourceTypes = scope.getSystemScope().getResourceTypes();
		while (iterator.hasNext()) {
			EObject curObject = iterator.next();
			for (EReference curReference : curObject.eClass().getEAllReferences()) {
				if (!curReference.isContainment() && !curReference.isContainer()) {
					if (ResourcetypePackage.Literals.RESOURCE_TYPE.isSuperTypeOf(curReference.getEReferenceType())) {
						ResourceType curType = (ResourceType) curObject.eGet(curReference);
						if (curType != null) {
							// find matching resources in system global scope
							// (matching is based on name and class
							for (ResourceType curPublicOne : publicResourceTypes.getResourceTypes()) {
								if (curType.eClass().isInstance(curPublicOne)
										&& curType.getName().equals(curPublicOne.getName())) {
									ModelSkeletonAdapter adapter = adapt(curType,
											ModelSkeletonAdapter.class);
									adapter.setRepositoryObject(curPublicOne);
									break;
								}
							}
						}
					}
				}
			}
		}
	}
}
