package tools.descartes.prisma.core.merge;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ObjectMatcher;
import tools.descartes.prisma.core.agent.AgentController;

public class ObjectMerger {

	private final AgentController agent;

	public ObjectMerger(AgentController agent) {
		this.agent = agent;
	}

	public void mergeTree(Transaction transaction, EObject base, EObject changed,
			Set<EStructuralFeature> changedFeatures) {
		// In order to optimize the performance we only consider features marked
		// as changed. If changedFeatures is null, all features are considered.
		Collection<EStructuralFeature> features = (changedFeatures == null)
				? changed.eClass().getEAllStructuralFeatures() : changedFeatures;
		for (EStructuralFeature curFeature : features) {
			if (curFeature instanceof EReference) {
				EReference curReference = (EReference) curFeature;
				if (curReference.isContainment()) {
					mergeFeature(transaction, base, changed, curFeature);
				}
			}
		}
	}

	public void mergeContents(Transaction transaction, EObject base, EObject changed,
			Set<EStructuralFeature> changedFeatures) {
		// In order to optimize the performance we only consider features marked
		// as changed. If changedFeatures is null, all features are considered.
		Collection<EStructuralFeature> features = (changedFeatures == null)
				? changed.eClass().getEAllStructuralFeatures() : changedFeatures;
		for (EStructuralFeature curFeature : features) {
			if (curFeature.isChangeable() && !curFeature.isDerived()) {
				if (curFeature instanceof EAttribute) {
					mergeFeature(transaction, base, changed, curFeature);
				} else if (curFeature instanceof EReference) {
					EReference curReference = (EReference) curFeature;
					// Containment references and their opposite container
					// references are merge in the mergeContainmentTree phase.
					if (!curReference.isContainment() && !curReference.isContainer()) {
						mergeFeature(transaction, base, changed, curFeature);
					}
				}
			}
		}
	}

	protected void mergeFeature(Transaction transaction, EObject base, EObject changed, EStructuralFeature curFeature) {
		if (changed.eIsSet(curFeature)) {
			if (curFeature.isMany()) {
				mergeManyValuesFeature(transaction, base, changed, curFeature);
			} else {
				mergeSingleValueFeature(transaction, base, changed, curFeature);
			}
		} else {
			if (base.eIsSet(curFeature)) {
				base.eUnset(curFeature);
			}
		}
	}

	protected void mergeSingleValueFeature(Transaction transaction, EObject base, EObject changed,
			EStructuralFeature feature) {
		Object oldValue = base.eGet(feature);
		Object newValue = changed.eGet(feature);
		Object mergeResult = oldValue;
		if (!areValuesEqual(feature, oldValue, newValue)) {
			mergeResult = mergeValue(feature, oldValue, newValue);
			base.eSet(feature, mergeResult);
		}
		connectValue(transaction, feature, newValue, mergeResult);
	}

	protected void connectValue(Transaction transaction, EStructuralFeature feature, Object skeletonObject,
			Object repositoryObject) {
		if ((feature instanceof EReference) && ((EReference) feature).isContainment()) {
			agent.getOwnershipManager().takeOwnership(transaction, (CDOObject) repositoryObject);

			ModelSkeletonAdapter valueAdapter = (ModelSkeletonAdapter) agent.adapt((EObject) skeletonObject,
					ModelSkeletonAdapter.class);
			valueAdapter.updateRepositoryObject((CDOObject) repositoryObject);
		}
	}

	protected void mergeManyValuesFeature(Transaction transaction, EObject base, EObject changed,
			EStructuralFeature feature) {
		@SuppressWarnings("unchecked")
		List<Object> baseList = (List<Object>) base.eGet(feature);
		List<?> changedList = (List<?>) changed.eGet(feature);

		if (!base.eIsSet(feature)) {
			// This feature has not been set before.
			// We can copy without checking for conflicts.
			for (Object curObject : changedList) {
				Object mergeResult = mergeValue(feature, null, curObject);
				baseList.add(mergeResult);
				connectValue(transaction, feature, curObject, mergeResult);
			}
		} else {
			// Perform a full search for all objects given that the order of the
			// lists may be different.
			for (Object curObject : changedList) {
				boolean contains = false;
				for (Object curBaseObject : baseList) {
					if (areValuesEqual(feature, curBaseObject, curObject)) {
						Object mergeResult = mergeValue(feature, curBaseObject, curObject);
						connectValue(transaction, feature, curObject, mergeResult);
						contains = true;
						break;
					}
				}
				if (!contains) {
					Object mergeResult = mergeValue(feature, null, curObject);
					baseList.add(mergeResult);
					connectValue(transaction, feature, curObject, mergeResult);
				}
			}
		}
	}

	protected Object mergeValue(EStructuralFeature feature, Object oldValue, Object newValue) {
		if (feature instanceof EAttribute) {
			return mergeAttributeValue(oldValue, newValue);
		} else {
			if (isContainment(feature)) {
				return mergeContainmentReferenceValue((EObject) oldValue, (EObject) newValue);
			} else {
				return mergeReferenceValue((EObject) oldValue, (EObject) newValue);
			}
		}
	}

	protected Object mergeAttributeValue(Object oldValue, Object newValue) {
		return newValue;
	}

	protected EObject mergeReferenceValue(EObject oldValue, EObject newValue) {
		ModelSkeletonAdapter valueAdapter = (ModelSkeletonAdapter) agent.adapt(newValue,
				ModelSkeletonAdapter.class);
		return valueAdapter.getTransientRepositoryObject();
	}

	private Object mergeContainmentReferenceValue(EObject oldValue, EObject newValue) {
		EObject mergedObject = oldValue;
		if (mergedObject == null) {
			mergedObject = EcoreUtil.create(newValue.eClass());
		}
		return mergedObject;
	}

	protected boolean areValuesEqual(EStructuralFeature feature, Object baseValue, Object changedValue) {
		if (baseValue == null || changedValue == null) {
			return false;
		}

		if (feature instanceof EAttribute) {
			return baseValue.equals(changedValue);
		} else {
			return ObjectMatcher.matches((EObject) baseValue, (EObject) changedValue);
		}
	}

	protected boolean isContainment(EStructuralFeature feature) {
		if (feature instanceof EReference) {
			return ((EReference) feature).isContainment();
		}
		return false;
	}
}
